<?php
/*
Plugin Name: DD_Roles
Version: 4.1
Plugin URI: http://dijkstradesign.com
Description: A plug-in to add and edit the roles and capabilities
Author: Wouter Dijkstra
Author URI: http://dijkstradesign.com
*/


/*  Copyright 2020  WOUTER DIJKSTRA  (email : info@wouterdijkstra.com)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License, version 2, as
    published by the Free Software Foundation.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


class dd_roles {

    const  defaultRoles = array('administrator','editor','author','contributor','subscriber','banned');

    private $groupedCapabilities;
    private $customCapabilities;
    private $defaultCapabilities;

    /**
     * dd_roles constructor.
     */
    public function __construct() {
        $this->setGroupedCapabilities();
        load_default_textdomain();
        register_activation_hook( __FILE__, array($this, 'activate_plugin') );
        register_deactivation_hook( __FILE__, array($this, 'deactivate_plugin') );
        add_action( 'admin_init', array( $this, 'dd_add_style_and_js_dd_roles' ) );
        add_action('admin_menu', array( $this, 'DD_rolesSettings'));
        add_filter('admin_footer_text', [ $this, 'replace_footer_admin']);
        add_action( 'wp_ajax_verify_and_add', array( $this, 'add_dd_role' ));
        add_action( 'wp_ajax_create_capability', array( $this, 'add_new_capabilities' ));
        add_action( 'wp_ajax_deleteRole', array( $this, 'deleteRole_callback' ));
        add_action( 'wp_ajax_migrateUsers', array( $this, 'migrateUsers_callback' ));
        add_action( 'wp_ajax_cleanUp', array( $this, 'cleanUp_callback' ));
        add_action( 'wp_ajax_changeCapState', array( $this, 'changeCapState_callback' ));
    }

    public function activate_plugin() {
        $this->add_dd_role('Banned');
    }

    public function deactivate_plugin(){
    }

    function replace_footer_admin () {
        global $pagenow;
        if($pagenow ==='users.php' && isset($_GET['page']) &&  $_GET['page'] === 'dd-roles.php' ) {
            remove_filter( 'update_footer', 'core_update_footer' );
            echo 'DD-roles is fully Fueled by <a href="http://www.wouterdijkstra.com" target="_blank">Wouter Dijkstra</a>. This tool is built on a passion for code and a love for Wordpress and free to use. <a href="https://www.paypal.com/cgi-bin/webscr?cmd=_donations&business=5V2C94HQAN63C&lc=US&item_name=Dijkstra%20Design&currency_code=EUR&bn=PP%2dDonationsBF%3abtn_donate_SM%2egif%3aNonHosted">Donating is a gift!</a>';
        }
    }

    /**
     *
     */
    public function dd_add_style_and_js_dd_roles()
    {
        if (current_user_can('administrator')) {
            wp_register_style( 'dd_roles_styles', plugins_url('/css/style.css', __FILE__) );
            wp_enqueue_style( 'dd_roles_styles' );
            wp_enqueue_script('select2', plugins_url('/js/select2.full.min.js', __FILE__), array('jquery'), '');
            wp_enqueue_script( 'dd_js_roles', plugins_url( '/js/default.min.js', __FILE__ ) , array( 'jquery','select2'), '' );
        }
    }

    function add_new_capabilities($new_capabilities = ''){

        $new_capabilities =  $new_capabilities ? $new_capabilities : $_POST['new_capability'];

        //array for feature: to clean up multi-select. For now 1 by 1
        $add_caps = array($new_capabilities);
        global $wp_roles;
        foreach ($add_caps as $cap) {
            foreach (array_keys($wp_roles->roles) as $role) {
                $wp_roles->add_cap($role, $cap);
                if ($role != 'administrator') {
                    $wp_roles->remove_cap( $role, $cap );
                }
            }
        }
    }

    function get_role_capability_counters($role){
        $admin_role = get_role('administrator');
        $admin_role_capabilities = array_keys( $admin_role->capabilities );
        $admin_role_total_active = count($admin_role_capabilities);
        $role = get_role($role);
        $role_capabilities = array_keys( $role->capabilities );
        $role_total_active = count($role_capabilities);
        $role_total_active = ($role_total_active) ? $role_total_active : 0;
        $percent = $role_total_active*(100/$admin_role_total_active);
        $percent = floor($percent).'%';

        $object = [
            "percent"=> $percent,
            "totalCapabilities"=> $admin_role_total_active,
            "totalRoleActiveCapabilities"=> $role_total_active,
        ];

        return $object;
    }

    function get_cap_count($role){
        $role = get_role($role);
        $rol_capabilities = array_keys( $role->capabilities );

        return count($rol_capabilities);
    }

    /**
     * @param $delCaps
     */
    function delete_capabilities($delCaps){
        $deleteCaps = is_array($delCaps) ? $delCaps : array($delCaps);
        global $wp_roles;
        foreach ($deleteCaps as $cap) {
            foreach (array_keys($wp_roles->roles) as $role) {
                $wp_roles->remove_cap($role, $cap);
            }
        }
    }

    function DD_rolesSettings() {

        add_users_page('DD Roles', 'DD Roles', 'manage_options', 'dd-roles.php', array($this,'rolesSettingsPage'));
        add_action( 'admin_init', array( $this, 'register_dd_roles_settings' ));
    }

    function rolesSettingsPage () {
        include_once('views/settingPage.php');
    }

    function getAllNonAdminRoles() {
        global $wp_roles;
        $roles = $wp_roles->get_names();
        unset($roles['administrator']);

        return $roles;
    }

    function onlyAdminRoles($onlyAdmin = true){
        foreach ($this->getAllNonAdminRoles() as $role_id => $role) {
            $totalUsers = self::getRoleTotalUsers($role_id);
            $onlyAdmin = $totalUsers ? false : $onlyAdmin;
        }
        return $onlyAdmin;
    }

    function deleteRole_callback() {
        $role_id = $_POST['role_id'];
        $this->userMigrate($role_id);

        $wp_roles = new WP_Roles();
        $wp_roles->remove_role($role_id);

        echo $role_id;
        die();
    }

    function migrateUsers_callback() {
        $fromRole = $_POST['fromRole'];
        $toRole = $_POST['toRole'];
        $this->userMigrate($fromRole,$toRole);

        die();
    }

    /**
     * @param $roleFrom
     * @param string $roleTo
     */
    function userMigrate($roleFrom, $roleTo='subscriber'){

        $usersRole = 'role='.$roleFrom.'';
        $users = get_users($usersRole);
        foreach ($users as $user) {
            $user->remove_role( $roleFrom );
            $user->add_role( $roleTo );
        }
    }

    /**
     * @param string $state
     * @param string $capName
     * @param string $role_id
     */
    function changeCapState_callback($state = '', $capName = '', $role_id = '') {
        $state = $state ? $state : filter_var( $_POST['state'], FILTER_VALIDATE_BOOLEAN);
        $capName = $capName ? $capName : $_POST['capName'];
        $role_id = $role_id ? $role_id : $_POST['role_id'];
        $role = get_role( $role_id);


        if($state){
            $role->add_cap( $capName );
            $object = json_encode( $this->get_role_capability_counters($role_id));
            echo $object;
        }
        else{
            $role->remove_cap( $capName );
            $object = json_encode(  $this->get_role_capability_counters($role_id));
            echo $object;
        }
        die();
    }

    function cleanUp_callback(){

        $delcaps = $_POST['delcaps'];

        //array for feature: to clean up multi-select. For now 1 by 1
        $delete_caps = array($delcaps);
        global $wp_roles;
        foreach ($delete_caps as $cap) {
            foreach (array_keys($wp_roles->roles) as $role) {
                $wp_roles->remove_cap($role, $cap);
            }
        }
        echo 'caps deleted';

        die();
    }

    function isRoleUnique($name) {
        global $wp_roles;
        $unique = true;
        $roleCredentials = $this->createRoleCredentials($name);
        $roles = isset($wp_roles) ? $wp_roles->get_names() : '';

        foreach ($roles as $role_id => $role_name) {
            if($roleCredentials[0] === $role_id){
                 $unique = false;
            }
        }
        //check if NOT alphanumeric....So don't pass
        $unique = !ctype_alnum($roleCredentials[0]) ? false : $unique;

        return $unique;
    }

    function createRoleCredentials($nameRaw) {
        $name = trim($nameRaw); //remove whitespace from begin and at end.
        $id = strtolower(str_replace(' ', '_', $name)); //replace whitespace in between, make lowercase.

        return array($id , $name);
    }

    function add_dd_role($new_displayName=''){

        $new_displayName = $new_displayName ? $new_displayName : $_POST['new_role_display_name'];
        $roleCredentials = $this->createRoleCredentials($new_displayName);
        $isUnique = $this->isRoleUnique($new_displayName);

        if ($isUnique){

            if ( ! isset( $GLOBALS[ 'wp_roles' ] ) ){
                $GLOBALS[ 'wp_roles' ] = new WP_Roles();
            }
            $GLOBALS[ 'wp_roles' ]->add_role( $roleCredentials[0], __( $new_displayName, 'cftp_band' ) );
        }

        if($new_displayName != 'Banned'){
            // callback for ajax
            echo $isUnique;
        }
    }

    function register_dd_roles_settings() {
        register_setting( 'dd_roles_settings', 'dd_roles' );
    }

    function setGroupedCapabilities() {
        $ordered_caps = [
            'read' => [
                'title' => "Dashboard, Profile and Update Core",
                "desc" => "Allows access to Administration Panel options \"Dashboard\" and \"Profile\" Allows access to the configuration parameters of widgets which included to the dashboard. Set permission to update WP-core",
                "filters" => ['edit_dashboard', 'update_core'],
                "capabilities" => []
            ],
            "posts" => [
                "title"=> "Posts",
                "desc"=> "Allows access to Administration Panel options \"Posts\".",
                "filters" => ['posts','manage_categories'],
                "capabilities"=>[]
            ],
            "pages" => [
                "title"=> "Pages",
                "desc"=> "Allows access to Administration Panel options \"Pages\".",
                "filters" => ['pages'],
                "capabilities" => []
            ],
            "comments"=> [
                "title"=> "Comments",
                "desc" => "Allows access to Administration Panel options \"Comments\" (\"Edit posts\" needs to be checked in Posts). To edit comments \"Edit others posts\" and \"Edit published posts\" need to be checked in Posts.",
                "filters" => ['comments'],
                "capabilities" => []
            ],
            "users"=> [
                "title"=> "Users",
                "desc"=> "Allows access to Administration Panel options \"Users\".",
                "filters" => ['users'],
                "capabilities"=> []
            ],
            "themes"=> [
                "title"=> "Appearance",
                "desc"=> "Allows access to Administration Panel options \"Appearance\".",
                "filters" => ['theme'],
                "capabilities"=> []
            ],
            "plugins"=> [
                "title"=> "Plugins",
                "desc"=> "Allows access to Administration Panel options \"Plugins\".",
                "filters" => ['plugins'],
                "capabilities"=> []
            ],
            "port"=> [
                "title"=> "Tools",
                "desc"=> "Allows access to Administration Panel options \"Tools\".",
                "filters" => ['port'],
                "capabilities"=> []
            ],
            "files"=> [
                "title"=> "Media",
                "desc"=> "Allows access to Administration Panel options \"Media\".",
                "filters" => ['files'],
                "capabilities"=> []
            ],
            "manage_options"=> [
                "title"=> "Settings",
                "desc"=> "Allows access to Administration Panel options \"Settings\".",
                "filters" => ['manage_options'],
                "capabilities"=> []
            ],
            "levels"=> [
                "title"=> "Deprecated, Levels & Rarely Used",
                "desc"=> "To maintain backwards compatibility with plugins that still use the user levels system (although this is very much discouraged), the default Roles in WordPress also include Capabilities that correspond to these levels. User Levels were finally deprecated in version 3.0.",
                "filters" => ['level', 'manage_links', 'unfiltered_html', 'unfiltered_upload'],
                "capabilities"=>[]
            ],
            "others"=> [
                "title" => 'Others',
                "desc" => 'all the other stuff!',
                "filters" => [],
                "capabilities" => []
            ]
        ];

        $default_caps = self::get_default_caps();
        $prefixes = array('assign_','edit_others_','edit_private_','edit_published_','edit_','publish_','read_private_','read_','delete_published_','delete_private_','delete_others_','delete_','manage_','_terms');
        $capa_types = [];

        foreach ($ordered_caps as $cap=>$params) {
            foreach( $default_caps as $capability ) {
                $isCorrectMapped = $capability === $cap;

                foreach($params['filters'] as $filter){
                    $isCorrectMapped = strpos($capability,$filter) !== false ? true : $isCorrectMapped;
                }
                if($isCorrectMapped) {
                    $this->defaultCapabilities[] = $capability;
                    $ordered_caps[$cap]['capabilities'][] = $capability;
                } else {
                    $ordered_caps['others']['capabilities'][] = $capability;
                }
            }

            // Order alphanum and Remove duplicates in 'others'
            sort($ordered_caps[$cap]['capabilities']);
            $ordered_caps['others']['capabilities'] = array_unique( array_diff($ordered_caps['others']['capabilities'], self::getAllGroupedCaps($ordered_caps)));
        }

        // loop through others capabilities en remove the prefixes
        foreach ($ordered_caps['others']['capabilities'] as $key=>$capability) {
            $raw_capa_type = str_replace($prefixes, "", $capability);
            $capa_types [] = $raw_capa_type;
        }
        // create unique values
        $capa_types = array_unique($capa_types);

        // remove the plurals
        foreach ($capa_types as $capa_type) {
            $plural = $capa_type.'s';
            if (($key = array_search($plural, $capa_types)) !== false) {
                unset($capa_types[$key]);
            }
        }

        foreach ($capa_types as $capa_type) {
            $name = ucfirst(str_replace('_', ' ', $capa_type));
            $ordered_caps[$capa_type]['title'] = $name;
            $ordered_caps[$capa_type]['desc'] = 'Allows access to capability type "'.$capa_type.'"';
            foreach ($ordered_caps['others']['capabilities'] as $capability) {

                if ($capability === $capa_type) {
                    $ordered_caps[$capa_type]['capabilities'][] = $capability;
                    $this->customCapabilities[$name][] = $capability;
                }
                foreach ($prefixes as $prefix) {
                    if( strpos($capability, $prefix.$capa_type) !== false || strpos($capability, $prefix.$capa_type.'s') !== false) {
                        $ordered_caps[$capa_type]['capabilities'][] = $capability;
                        $this->customCapabilities[$name][] = $capability;
                    }
                }
            }
        }
        // remove 'others'
        unset($ordered_caps['others']);
        $this->groupedCapabilities = $ordered_caps;
    }

    /**
     * Get all capabilities grouped UX friendly
     * @return array
     */
    function getGroupedCapabilities(){
        return $this->groupedCapabilities;
    }

    function get_default_caps(){
        $role =  get_role( 'administrator' );
        return array_keys($role->capabilities);
    }

    /**
     * @param $role_id
     * @return int
     */
    function getRoleTotalUsers($role_id){
        $result = count_users();
        $availRoles = $result['avail_roles'];
        $roleTotalUsers = array_key_exists($role_id, $availRoles) ? $availRoles[$role_id] : 0;

        return (int)$roleTotalUsers;
    }

    function get_caps_names($role_id){

        $role_id = get_role($role_id);
        return array_keys( $role_id->capabilities );
    }

    /**
     * Gives array of all grouped capabilities
     * @param $ordered_caps
     * @return array
     */
    function getAllGroupedCaps($ordered_caps) {
        $caps = [];

        foreach ($ordered_caps as $group=>$params){
            if($group === 'others'):break; endif;
            foreach ($params['capabilities'] as $capability) {
                $caps[] = $capability;
            }
        }
        return $caps;
    }

    function getRoles() {
        global $wp_roles;
        $roles = $wp_roles->get_names();

        // move 'banned' to end
        $bannedRole = $roles['banned'];
        unset($roles['banned']);
        $roles['banned'] = $bannedRole;

        return $roles;
    }
}

$dd_roles = new dd_roles();